from __future__ import annotations

import asyncio
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Collection,
    Deque,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

A = TypeVar("A")
T = TypeVar("T")
R = TypeVar("R")


async def consume(iterable: AsyncIterable[Any]) -> None:
    async for _ in iterable:
        pass


class BoostExecutor:
    """BoostExecutor implements a concurrent.futures-like interface for running async tasks.

    The idea is: you want to run a limited number of tasks at a time. Moreover, if you have spare
    capacity, you want to redistribute it to tasks you're running. For instance, when downloading
    chunks of a file, we could use spare capacity to proactively download future chunks. That is,
    instead of the following serial code:
    ```
    for chunk in iterable_of_chunks:
        data = await download_func(chunk)
        write(data)
    ```
    You can instead do:
    ```
    async with BoostExecutor(4) as executor:
        async for data in executor.map_ordered(download_func, iterable_of_chunks):
            write(data)
    ```

    The number supplied to BoostExecutor is the number of units of concurrency you wish to maintain.
    That is, using ``BoostExecutor(1)`` would make the above example function approximately similar
    to the serial code.

    """

    def __init__(self, concurrency: int) -> None:
        assert concurrency > 0
        self.concurrency = concurrency
        # Okay, so this is a little tricky. We take away one unit of concurrency now, and give it
        # back when you iterate over a boostable. You can think of concurrency - 1 as the number of
        # units of "background" concurrency. When our "foreground" starts iterating over a
        # boostable, for instance, with an async for loop, it donates one unit of concurrency to the
        # cause for the duration of the iteration.
        # Why this rigmarole? To avoid deadlocks on reentrancy. If a boostable spawns more work on
        # the same executor and then blocks on that work, it's holding a unit of concurrency while
        # trying to acquire another one. If that happens enough times, we deadlock. For examples,
        # see test_composition_nested_(un)?ordered.
        # To solve this, we treat Boostable.__aiter__ as an indicator that we are entrant. We then
        # donate a unit of concurrency by releasing the semaphore. Conceptually, this is our
        # "foreground" or "re-entrant" unit of concurrency. We give this back when the Boostable is
        # done iterating. We could be more granular and do this on every Boostable.__anext__, but
        # that's a lot more overhead, makes iteration less predictable and can result in what feels
        # like unnecessary blocking. We could also do something complicated using contextvars, but
        # that would have to be something complicated and would have to use contextvars.
        self.semaphore = asyncio.Semaphore(concurrency - 1)
        self.boostables: Deque[Boostable[Any]] = Deque()

        self.waiter: Optional[asyncio.Future[None]] = None
        self.runner: Optional[asyncio.Task[None]] = None
        self.shutdown: bool = False

    async def __aenter__(self) -> BoostExecutor:
        self.runner = asyncio.create_task(self.run())
        return self

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_value: Any, traceback: Any
    ) -> None:
        self.shutdown = True
        if exc_type:
            # If there was an exception, let it propagate and don't block on the runner exiting.
            # Also cancel the runner.
            assert self.runner is not None
            self.runner.cancel()
            return
        self.notify_runner()
        assert self.runner is not None
        await self.runner

    def eagerise(self, iterator: AsyncIterator[T]) -> EageriseBoostable[T]:
        ret = EageriseBoostable(iterator, self)
        self.boostables.appendleft(ret)
        self.notify_runner()
        return ret

    def map_ordered(
        self, func: Callable[[A], Awaitable[T]], iterable: BoostUnderlying[A]
    ) -> OrderedMappingBoostable[A, T]:
        ret = OrderedMappingBoostable(func, iterable, self)
        self.boostables.appendleft(ret)
        self.notify_runner()
        return ret

    def map_unordered(
        self, func: Callable[[A], Awaitable[T]], iterable: BoostUnderlying[A]
    ) -> UnorderedMappingBoostable[A, T]:
        ret = UnorderedMappingBoostable(func, iterable, self)
        self.boostables.appendleft(ret)
        self.notify_runner()
        return ret

    def enumerate(self, iterable: BoostUnderlying[T]) -> EnumerateBoostable[T]:
        ret = EnumerateBoostable(iterable, self)
        self.boostables.appendleft(ret)
        self.notify_runner()
        return ret

    def filter(
        self, filter_fn: Optional[Callable[[T], bool]], iterable: BoostUnderlying[T]
    ) -> FilterBoostable[T]:
        ret = FilterBoostable(filter_fn, iterable, self)
        self.boostables.appendleft(ret)
        self.notify_runner()
        return ret

    def notify_runner(self) -> None:
        if self.waiter and not self.waiter.done():
            self.waiter.set_result(None)

    async def run(self) -> None:
        loop = asyncio.get_running_loop()
        exhausted_boostables: List[Boostable[Any]] = []
        not_ready_boostables: Deque[Boostable[Any]] = Deque()

        MIN_TIMEOUT = 0.01
        MAX_TIMEOUT = 0.1
        timeout = MIN_TIMEOUT

        if self.concurrency == 1:
            return

        while True:
            await self.semaphore.acquire()
            self.semaphore.release()

            while self.boostables:
                # We round robin the boostables until they're either all exhausted or not ready
                task = self.boostables[0].provide_boost()
                if isinstance(task, NotReady):
                    not_ready_boostables.append(self.boostables.popleft())
                    continue
                if isinstance(task, Exhausted):
                    exhausted_boostables.append(self.boostables.popleft())
                    continue
                assert isinstance(task, asyncio.Task)

                await asyncio.sleep(0)
                self.boostables.rotate(-1)
                if self.semaphore.locked():
                    break
            else:
                self.boostables = not_ready_boostables
                not_ready_boostables = Deque()

            if self.semaphore.locked():
                # If we broke out of the inner loop due to a lack of available concurrency, go to
                # the top of the outer loop and wait for concurrency
                continue

            if self.shutdown and not self.boostables:
                # If we've been told to shutdown and we have nothing more to boost, exit
                break

            self.waiter = loop.create_future()
            try:
                # If we still have boostables (because they're currently not ready), timeout in
                # case they become ready
                # Otherwise, (since we have available concurrency), wait for us to be notified in
                # case of more work
                await asyncio.wait_for(self.waiter, timeout if self.boostables else None)
                # If we were waiting for boostables to become ready, increase how long we'd wait
                # the next time. Otherwise, reset the timeout duration.
                timeout = min(MAX_TIMEOUT, timeout * 2) if self.boostables else MIN_TIMEOUT
            except asyncio.TimeoutError:
                pass
            self.waiter = None

        # It's somewhat unintuitive that we can shutdown an executor while Boostables are still
        # being iterated over. So wait for them to finish as a courtesy (but not a guarantee).
        for boostable in exhausted_boostables:
            await boostable.wait()

        # Yield, so that iterations over boostables are likely to have finished by the time we exit
        # the BoostExecutor context
        await asyncio.sleep(0)


class Exhausted:
    pass


class NotReady:
    pass


class Boostable(Generic[T]):
    """A Boostable is an async iterable with a twist.

    The twist is that it can make use of additional concurrency to compute the elements it iterates
    in parallel. Boostables also compose, so that additional concurrency is directed to the task
    that needs it most.

    See the docstring of MappingBoostable to see how this plays out in practice.

    You probably shouldn't use these classes directly, instead, instantiate them via the
    corresponding methods of the executor used to manage the Boostable.

    """

    def __init__(self, executor: BoostExecutor) -> None:
        self.executor = executor

    def provide_boost(self) -> Union[NotReady, Exhausted, asyncio.Task[Any]]:
        """Start an asyncio task to help speed up this boostable.

        Returns NotReady if we can't make use of a boost currently.
        Returns Exhausted if we're done. This causes the BoostExecutor to stop attempting to provide
        boosts to this Boostable.

        """
        raise NotImplementedError

    async def wait(self) -> None:
        """Wait on executor shutdown, if helpful.

        Waiting if we still have work pending makes executor shutdown more intuitive.
        Boostables don't really make any guarantees about their behaviour here, though.

        """
        pass

    def dequeue(self) -> Union[NotReady, Exhausted, T]:
        """Non-blockingly dequeue a result we have ready.

        Returns NotReady if it's not ready to dequeue.
        Returns Exhausted if we're definitely done.

        """
        raise NotImplementedError

    async def blocking_dequeue(self) -> T:
        """Dequeue a result, waiting if necessary.

        Raises StopAsyncIteration if exhausted.

        """
        raise NotImplementedError

    def __aiter__(self) -> AsyncIterator[T]:
        async def iterator() -> AsyncIterator[T]:
            try:
                self.executor.semaphore.release()
                while True:
                    yield await self.blocking_dequeue()
            except StopAsyncIteration:
                pass
            finally:
                await self.executor.semaphore.acquire()

        return iterator()


class MappingBoostable(Boostable[T], Generic[A, T]):
    """Represents a map over an underlying iterable.

    MappingBoostables map an async function over each element of some underlying iterable. To boost
    this operation, we dequeue another element from the underlying iterable and create an asyncio
    task to perform the async function call over that element.

    MappingBoostables also compose; that is, the underlying iterable can be another Boostable. In
    this case, we quickly (aka non-blockingly) check whether the underlying Boostable has an element
    ready. If so, we dequeue it, just as we would with an iterable, if not, we forward the boost we
    were provided to the underlying Boostable.

    MappingBoostables come in two flavours: ordered and unordered. Remember that Boostables are
    async iterables. An OrderedMappingBoostable will yield results corresponding to the order of the
    underlying iterable, whereas an UnorderedMappingBoostable will yield results based on what
    finishes first.

    Note that both kinds of MappingBoostables will, of course, dequeue from the underlying iterable
    in whatever order the underlying provides. This means we'll start tasks in order (but not
    necessarily finish them in order, which is kind of the point).
    """

    buffer: Collection[Awaitable[T]]

    def __init__(
        self,
        func: Callable[[A], Awaitable[T]],
        iterable: BoostUnderlying[A],
        executor: BoostExecutor,
    ) -> None:
        super().__init__(executor)

        if not isinstance(iterable, (Iterator, Boostable)):
            raise ValueError("Underlying iterable must be an Iterator or Boostable")

        async def wrapper(arg: A) -> T:
            async with self.executor.semaphore:
                return await func(arg)

        self.func = wrapper
        self.iterable = iterable

    async def wait(self) -> None:
        if self.buffer:
            await asyncio.wait(self.buffer)

    def provide_boost(self) -> Union[NotReady, Exhausted, asyncio.Task[Any]]:
        if not self.executor.shutdown and len(self.buffer) > 2 * self.executor.concurrency:
            # if we have a lot of stuff ready to go, apply backpressure by not accepting the boost
            # the main effect of this is to reduce memory usage
            # always accept boosts if the executor is shutting down to prevent hangs on misuse
            return NotReady()

        result = dequeue_underlying(self.iterable)
        if isinstance(result, NotReady):
            if isinstance(self.iterable, Boostable):
                # the underlying boostable doesn't have anything ready for us
                # so forward the boost to the underlying boostable
                return self.iterable.provide_boost()
            return result
        if isinstance(result, Exhausted):
            return result

        return self.enqueue(result)

    def enqueue(self, arg: A) -> asyncio.Task[T]:
        """Start and return an asyncio task based on an element from the underlying."""
        raise NotImplementedError


class OrderedMappingBoostable(MappingBoostable[A, T]):
    def __init__(
        self,
        func: Callable[[A], Awaitable[T]],
        iterable: BoostUnderlying[A],
        executor: BoostExecutor,
    ) -> None:
        super().__init__(func, iterable, executor)
        self.buffer: Deque[asyncio.Task[T]] = Deque()

    def enqueue(self, arg: A) -> asyncio.Task[T]:
        task = asyncio.create_task(self.func(arg))
        self.buffer.append(task)
        return task

    def dequeue(self) -> Union[NotReady, T]:
        if not self.buffer or not self.buffer[0].done():
            return NotReady()
        task = self.buffer.popleft()
        return task.result()

    async def blocking_dequeue(self) -> T:
        while True:
            if not self.buffer:
                arg = await blocking_dequeue_underlying(self.iterable)
                self.enqueue(arg)
            ret = self.dequeue()
            if not isinstance(ret, NotReady):
                return ret
            # note that dequeues are racy, so we can't return the result of self.buffer[0] instead,
            # we just take it as a signal that something is likely ready, and loop and try to
            # dequeue
            await self.buffer[0]


class UnorderedMappingBoostable(MappingBoostable[A, T]):
    def __init__(
        self,
        func: Callable[[A], Awaitable[T]],
        iterable: BoostUnderlying[A],
        executor: BoostExecutor,
    ) -> None:
        super().__init__(func, iterable, executor)
        self.buffer: Set[asyncio.Task[T]] = set()
        self.waiter: Optional[asyncio.Future[asyncio.Task[T]]] = None

    def done_callback(self, task: asyncio.Task[T]) -> None:
        if self.waiter and not self.waiter.done():
            self.waiter.set_result(task)

    def enqueue(self, arg: A) -> asyncio.Task[T]:
        task = asyncio.create_task(self.func(arg))
        self.buffer.add(task)
        task.add_done_callback(self.done_callback)
        return task

    def dequeue(self, hint: Optional[asyncio.Task[T]] = None) -> Union[NotReady, T]:
        # hint is a task that we suspect is dequeuable, which allows us to skip the linear check
        # against all outstanding tasks in the happy case.
        if hint is not None and hint in self.buffer and hint.done():
            task = hint
        else:
            try:
                task = next(t for t in self.buffer if t.done())
            except StopIteration:
                return NotReady()
        self.buffer.remove(task)
        return task.result()

    async def blocking_dequeue(self) -> T:
        task = None
        loop = asyncio.get_running_loop()
        while True:
            if not self.buffer:
                arg = await blocking_dequeue_underlying(self.iterable)
                task = self.enqueue(arg)
            ret = self.dequeue(hint=task)
            if not isinstance(ret, NotReady):
                return ret
            # note that dequeues are racy, so the task we get from self.waiter may already have been
            # dequeued. in the happy case, however, it's probably the task we want to return, so we
            # use it as the hint for the next dequeue.
            self.waiter = loop.create_future()
            task = await self.waiter
            self.waiter = None


class FilterBoostable(Boostable[T]):
    def __init__(
        self,
        filter_fn: Optional[Callable[[T], bool]],
        inner: BoostUnderlying[T],
        executor: BoostExecutor,
    ):
        super().__init__(executor)
        self.filter_fn = filter_fn or bool
        self.inner = inner

    def provide_boost(self) -> Union[NotReady, Exhausted, asyncio.Task[Any]]:
        if isinstance(self.inner, Boostable):
            return self.inner.provide_boost()
        return Exhausted()

    def dequeue(self) -> Union[NotReady, Exhausted, T]:
        while True:
            ret = dequeue_underlying(self.inner)
            if isinstance(ret, (NotReady, Exhausted)) or self.filter_fn(ret):
                return ret

    async def blocking_dequeue(self) -> T:
        while True:
            ret = await blocking_dequeue_underlying(self.inner)
            if self.filter_fn(ret):
                return ret


class EnumerateBoostable(Boostable[Tuple[int, T]]):
    def __init__(self, inner: BoostUnderlying[T], executor: BoostExecutor):
        super().__init__(executor)
        self.inner = inner
        self.index = 0

    def provide_boost(self) -> Union[NotReady, Exhausted, asyncio.Task[Any]]:
        if isinstance(self.inner, Boostable):
            return self.inner.provide_boost()
        return Exhausted()

    def dequeue(self) -> Union[NotReady, Exhausted, Tuple[int, T]]:
        inner_ret = dequeue_underlying(self.inner)
        if isinstance(inner_ret, (NotReady, Exhausted)):
            return inner_ret
        ret = (self.index, inner_ret)
        self.index += 1
        return ret

    async def blocking_dequeue(self) -> Tuple[int, T]:
        inner_ret = await blocking_dequeue_underlying(self.inner)
        ret = (self.index, inner_ret)
        self.index += 1
        return ret


class EageriseBoostable(Boostable[T]):
    def __init__(self, iterable: AsyncIterator[T], executor: BoostExecutor) -> None:
        super().__init__(executor)
        self.iterable = iterable
        self.buffer: Deque[asyncio.Task[T]] = Deque()
        self.done: bool = False

        self.waiter_buffer: Optional[asyncio.Future[None]] = None
        self.waiter_backpressure: Optional[asyncio.Future[None]] = None

        self.buffer_task = asyncio.create_task(self.eagerly_buffer())

    def provide_boost(self) -> Union[NotReady, Exhausted, asyncio.Task[Any]]:
        # If we always returned Exhausted, that would cause MappingBoostable to in turn return
        # Exhausted, if it couldn't currently make use of a boost.
        return Exhausted() if self.done else NotReady()

    async def wait(self) -> None:
        await self.buffer_task

    def dequeue(self) -> Union[NotReady, Exhausted, T]:
        if not self.buffer:
            return Exhausted() if self.done else NotReady()

        task = self.buffer.popleft()

        if self.waiter_backpressure:
            self.waiter_backpressure.set_result(None)
            self.waiter_backpressure = None

        return task.result()

    async def blocking_dequeue(self) -> T:
        loop = asyncio.get_running_loop()
        while True:
            ret = self.dequeue()
            if isinstance(ret, Exhausted):
                raise StopAsyncIteration
            if not isinstance(ret, NotReady):
                return ret

            self.waiter_buffer = loop.create_future()
            await self.waiter_buffer

    async def eagerly_buffer(self) -> None:
        loop = asyncio.get_running_loop()
        async with self.executor.semaphore:
            # We can't use async for because we need to preserve exceptions
            it = self.iterable.__aiter__()
            while True:
                # I guess in theory, __anext__ isn't guaranteed to be a coroutine
                # https://github.com/hauntsaninja/boostedblob/pull/12
                task: asyncio.Task[T] = asyncio.create_task(it.__anext__())  # type: ignore [arg-type]
                try:
                    await task
                except StopAsyncIteration:
                    break
                except Exception:
                    pass
                self.buffer.append(task)

                if self.waiter_buffer:
                    self.waiter_buffer.set_result(None)
                    self.waiter_buffer = None

                # apply backpressure
                if not self.executor.shutdown and len(self.buffer) > 10 * self.executor.concurrency:
                    self.executor.semaphore.release()
                    self.waiter_backpressure = loop.create_future()
                    await self.waiter_backpressure
                    await self.executor.semaphore.acquire()

            self.done = True
            if self.waiter_buffer:
                self.waiter_buffer.set_result(None)
                self.waiter_buffer = None


# see docstring of Boostable
BoostUnderlying = Union[Iterator[T], Boostable[T]]


def dequeue_underlying(iterable: BoostUnderlying[T]) -> Union[NotReady, Exhausted, T]:
    """Like ``dequeue``, but abstracts over a BoostUnderlying."""
    if isinstance(iterable, Boostable):
        return iterable.dequeue()
    if isinstance(iterable, Iterator):
        try:
            return next(iterable)
        except StopIteration:
            return Exhausted()
    raise AssertionError


async def blocking_dequeue_underlying(iterable: BoostUnderlying[T]) -> T:
    """Like ``blocking_dequeue``, but abstracts over a BoostUnderlying."""
    if isinstance(iterable, Boostable):
        return await iterable.blocking_dequeue()
    if isinstance(iterable, Iterator):
        try:
            return next(iterable)
        except StopIteration as e:
            raise StopAsyncIteration from e
    raise AssertionError


async def iter_underlying(iterable: BoostUnderlying[T]) -> AsyncIterator[T]:
    """Like ``iter``, but abstracts over a BoostUnderlying."""
    if isinstance(iterable, Boostable):
        async for x in iterable:
            yield x
    elif isinstance(iterable, Iterator):
        for x in iterable:
            yield x
    else:
        raise AssertionError



from __future__ import annotations

import asyncio
import base64
import hashlib
import os
import random
from typing import List, Mapping, Optional, Tuple, Union

from .boost import BoostExecutor, BoostUnderlying, consume, iter_underlying
from .delete import remove
from .globals import config
from .path import AzurePath, BasePath, CloudPath, GooglePath, LocalPath, exists, pathdispatch
from .read import ByteRange
from .request import (
    Request,
    RequestFailure,
    azurify_request,
    exponential_sleep_generator,
    googlify_request,
)
from .xml import etree

AZURE_BLOCK_COUNT_LIMIT = 50_000

# ==============================
# write_single
# ==============================


@pathdispatch
async def write_single(path: Union[BasePath, str], data: bytes, overwrite: bool = False) -> None:
    """Write the given stream to ``path``.

    :param path: The path to write to.
    :param data: The data to write.
    :param overwrite: If False, raises if the path already exists.

    """
    raise ValueError(f"Unsupported path: {path}")


@write_single.register  # type: ignore
async def _azure_write_single(path: AzurePath, data: bytes, overwrite: bool = False) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            data=data,
            headers={"x-ms-blob-type": "BlockBlob"},
            success_codes=(201,),
        )
    )
    await request.execute_reponseless()


@write_single.register  # type: ignore
async def _google_write_single(path: GooglePath, data: bytes, overwrite: bool = False) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    request = await googlify_request(
        Request(
            method="POST",
            url=path.format_url(
                "https://storage.googleapis.com/upload/storage/v1/b/{bucket}/o?uploadType=media&name={blob}"
            ),
            data=data,
            headers={"Content-Type": "application/octet-stream"},
        )
    )
    await request.execute_reponseless()


@write_single.register  # type: ignore
async def _local_write_single(path: LocalPath, data: bytes, overwrite: bool = False) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    os.makedirs(path.parent, exist_ok=True)
    with open(path, mode="wb") as f:
        f.write(data)


# ==============================
# write_stream
# ==============================


@pathdispatch
async def write_stream(
    path: Union[BasePath, str],
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    """Write the given stream to ``path``.

    :param path: The path to write to.
    :param executor: An executor.
    :param stream: The stream of bytes to write. Note the chunking of stream also determines how
        we chunk the writes. Writes to Google Cloud must be chunked in multiples of 256 KB.
    :param overwrite: If False, raises if the path already exists.

    """
    raise ValueError(f"Unsupported path: {path}")


@write_stream.register  # type: ignore
async def _azure_write_stream(
    path: AzurePath,
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    if overwrite:
        await prepare_block_blob_write(path)
    else:
        if await exists(path):
            raise FileExistsError(path)

    upload_id = get_upload_id()
    md5 = hashlib.md5()
    max_block_index = -1

    async def upload_chunk(index_chunk: Tuple[int, bytes]) -> None:
        block_index, chunk = index_chunk
        # https://docs.microsoft.com/en-us/rest/api/storageservices/put-block-list#remarks
        assert block_index < AZURE_BLOCK_COUNT_LIMIT

        nonlocal max_block_index
        max_block_index = max(max_block_index, block_index)

        block_id = get_block_id(upload_id, block_index)
        md5.update(chunk)

        await _azure_put_block(path, block_id, chunk)

    await consume(executor.map_ordered(upload_chunk, executor.enumerate(stream)))

    # azure does not calculate md5s for us, we have to do that manually
    # https://blogs.msdn.microsoft.com/windowsazurestorage/2011/02/17/windows-azure-blob-md5-overview/
    headers = {"x-ms-blob-content-md5": base64.b64encode(md5.digest()).decode("utf8")}
    blocklist = [get_block_id(upload_id, i) for i in range(max_block_index + 1)]
    await azure_put_block_list(path, blocklist, headers=headers)


@write_stream.register  # type: ignore
async def _google_write_stream(
    path: GooglePath,
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    upload_url = await _google_start_resumable_upload(path)

    offset = 0
    is_finalised = False
    total_size = "*"
    # Unfortunately, GCS doesn't allow us to upload these chunks in parallel.
    async for chunk in iter_underlying(stream):
        start = offset
        offset += len(chunk)
        end = offset

        # GCS requires resumable uploads to be chunked in multiples of 256 KB, except for the last
        # chunk. If you upload a chunk with another size you get an HTTP 400 error, unless you tell
        # GCS that it's the last chunk. Since our interface doesn't allow us to know whether or not
        # a given chunk is actually the last chunk, we go ahead and assume that it is if it's an
        # invalid chunk size. If we receive multiple chunks of invalid chunk size, we throw an
        # error.
        should_finalise = len(chunk) % (256 * 1024) != 0
        if should_finalise:
            if is_finalised:
                raise ValueError(
                    "The upload was already finalised. A likely cause is the given stream was "
                    "chunked incorrectly. Uploads to Google Cloud need to be chunked in "
                    "multiples of 256 KB (except for the last chunk)."
                )
            total_size = str(end)
            is_finalised = True

        request = await googlify_request(
            Request(
                method="PUT",
                url=upload_url,
                data=chunk,
                headers={
                    "Content-Type": "application/octet-stream",
                    "Content-Range": f"bytes {start}-{end-1}/{total_size}",
                },
                success_codes=(200, 201) if is_finalised else (308,),
            )
        )
        await request.execute_reponseless()

    if not is_finalised:
        await _google_finalise_upload(upload_url, total_size=offset)


@write_stream.register  # type: ignore
async def _local_write_stream(
    path: LocalPath,
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    os.makedirs(path.parent, exist_ok=True)
    loop = asyncio.get_running_loop()
    with open(path, mode="wb") as f:
        async for data in iter_underlying(stream):
            await loop.run_in_executor(config.executor, f.write, data)


# ==============================
# write_stream_unordered
# ==============================


@pathdispatch
async def write_stream_unordered(
    path: Union[CloudPath, str],
    stream: BoostUnderlying[Tuple[bytes, ByteRange]],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    """Write the given stream to ``path``.

    :param path: The path to write to.
    :param executor: An executor.
    :param stream: The stream of bytes to write, along with what range of bytes each chunk
        corresponds to. Note the chunking of stream also determines how we chunk the writes.
    :param overwrite: If False, raises if the path already exists.

    """
    raise ValueError(f"Unsupported path: {path}")


@write_stream_unordered.register  # type: ignore
async def _azure_write_stream_unordered(
    path: AzurePath,
    stream: BoostUnderlying[Tuple[bytes, ByteRange]],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    # TODO: this doesn't upload an md5...
    if overwrite:
        await prepare_block_blob_write(path)
    else:
        if await exists(path):
            raise FileExistsError(path)

    upload_id = get_upload_id()
    block_list = []

    async def upload_chunk(index_chunk_byte_range: Tuple[int, Tuple[bytes, ByteRange]]) -> None:
        unordered_index, (chunk, byte_range) = index_chunk_byte_range
        # https://docs.microsoft.com/en-us/rest/api/storageservices/put-block-list#remarks
        assert unordered_index < AZURE_BLOCK_COUNT_LIMIT

        block_id = get_block_id(upload_id, unordered_index)
        block_list.append((byte_range[0], unordered_index))

        await _azure_put_block(path, block_id, chunk)

    await consume(executor.map_unordered(upload_chunk, executor.enumerate(stream)))

    # sort by start byte so the blocklist is ordered correctly
    block_list.sort()
    await azure_put_block_list(path, [get_block_id(upload_id, index) for _, index in block_list])


@write_stream_unordered.register  # type: ignore
async def _google_write_stream_unordered(
    path: GooglePath,
    stream: BoostUnderlying[Tuple[bytes, ByteRange]],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    raise NotImplementedError


# ==============================
# helpers
# ==============================


def get_upload_id() -> int:
    return random.randint(0, 2**47 - 1)


def get_block_id(upload_id: int, index: int) -> str:
    assert index < 2**17
    id_plus_index = (upload_id << 17) + index
    assert id_plus_index < 2**64
    return base64.b64encode(id_plus_index.to_bytes(8, byteorder="big")).decode("utf8")


async def prepare_block_blob_write(path: AzurePath, _always_clear: bool = False) -> None:
    request = await azurify_request(
        Request(
            method="GET",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            params=dict(comp="blocklist"),
            success_codes=(200, 404, 400),
        )
    )
    async with request.execute() as resp:
        # If the block doesn't exist, we're good to go.
        if resp.status == 404:
            return
        # Get Block List will return 400 if the blob is not a block blob.
        # In this case, we need to remove the blob to prevent errors when we Put Block
        if resp.status == 400:
            await remove(path)
            return
        data = await resp.read()

    result = etree.fromstring(data)
    blocks = [b.text for b in result.iterfind("CommittedBlocks/Block/Name")]
    if not blocks:
        return

    if not _always_clear and len(blocks) < 10_000:
        # Don't bother clearing uncommitted blocks if we have less than 10,000 committed blocks,
        # since seems unlikely we'd run into the 100,000 block limit in this case.
        return

    # In the event of multiple concurrent writers, we run the risk of ending up with uncommitted
    # blocks which could hit the uncommited block limit. Clear uncommitted blocks by performing a
    # Put Block List with all the existing blocks. This will change the last-modified timestamp and
    # the etag.

    # Make sure to preserve metadata for the file
    request = await azurify_request(
        Request(
            method="HEAD",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    async with request.execute() as resp:
        metadata = resp.headers
    headers = {k: v for k, v in metadata.items() if k.startswith("x-ms-meta-")}
    RESPONSE_HEADER_TO_REQUEST_HEADER = {
        "Cache-Control": "x-ms-blob-cache-control",
        "Content-Type": "x-ms-blob-content-type",
        "Content-MD5": "x-ms-blob-content-md5",
        "Content-Encoding": "x-ms-blob-content-encoding",
        "Content-Language": "x-ms-blob-content-language",
        "Content-Disposition": "x-ms-blob-content-disposition",
    }
    for src, dst in RESPONSE_HEADER_TO_REQUEST_HEADER.items():
        if src in metadata:
            headers[dst] = metadata[src]

    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            headers={**headers, "If-Match": metadata["etag"]},
            params=dict(comp="blocklist"),
            data={"BlockList": {"Latest": blocks}},
            success_codes=(201, 404, 412),
        )
    )
    await request.execute_reponseless()


async def _azure_put_block(path: AzurePath, block_id: str, chunk: bytes) -> None:
    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            params=dict(comp="block", blockid=block_id),
            data=chunk,
            success_codes=(201,),
        )
    )
    await request.execute_reponseless()


async def azure_put_block_list(
    path: AzurePath, block_list: List[str], headers: Optional[Mapping[str, str]] = None
) -> None:
    if headers is None:
        headers = {}
    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            headers=headers,
            params=dict(comp="blocklist"),
            data={"BlockList": {"Latest": block_list}},
            success_codes=(201, 400),
        )
    )

    for attempt, backoff in enumerate(
        exponential_sleep_generator(
            initial=config.backoff_initial,
            maximum=config.backoff_max,
            jitter_fraction=config.backoff_jitter_fraction,
        )
    ):
        async with request.execute() as resp:
            if resp.status == 201:
                return

            assert resp.status == 400
            data = await resp.read()
            result = etree.fromstring(data)
            if result.findtext("Code") == "InvalidBlockList":
                if attempt >= 3:
                    raise RuntimeError(
                        f"Invalid block list, most likely due to concurrent writes to {path}"
                    )
                # It turns out these can actually succeed on retry.
                # We run into this when using unordered upload, maybe it means we're just
                # going too fast...
                await asyncio.sleep(backoff)
                continue

            raise RequestFailure(reason=str(resp.reason), request=request, status=resp.status)


async def _google_start_resumable_upload(path: GooglePath) -> str:
    request = await googlify_request(
        Request(
            method="POST",
            url=path.format_url(
                "https://storage.googleapis.com/upload/storage/v1/b/{bucket}/o?uploadType=resumable"
            ),
            data=dict(name=path.blob),
            headers={"Content-Type": "application/json; charset=UTF-8"},
            failure_exceptions={400: FileNotFoundError(), 404: FileNotFoundError()},
        )
    )
    async with request.execute() as resp:
        upload_url = resp.headers["Location"]
        return upload_url


async def _google_finalise_upload(upload_url: str, total_size: int) -> None:
    headers = {"Content-Type": "application/octet-stream", "Content-Range": f"bytes */{total_size}"}
    request = await googlify_request(
        Request(method="PUT", url=upload_url, headers=headers, success_codes=(200, 201))
    )
    await request.execute_reponseless()



from __future__ import annotations

import base64
import datetime
import hmac
import json
import os
import re
import time
import urllib.parse
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, Tuple

from .xml import etree

if TYPE_CHECKING:
    from .path import AzurePath
    from .request import Request


AZURE_SAS_TOKEN_EXPIRATION_SECONDS = 60 * 60
# these seem to be expired manually, but we don't currently detect that
AZURE_SHARED_KEY_EXPIRATION_SECONDS = 24 * 60 * 60
OAUTH_TOKEN = "oauth_token"
SHARED_KEY = "shared_key"


def load_credentials() -> Dict[str, Any]:
    # AZURE_STORAGE_KEY seems to be the environment variable mentioned by the az cli
    # AZURE_STORAGE_ACCOUNT_KEY is mentioned elsewhere on the internet
    for varname in ["AZURE_STORAGE_KEY", "AZURE_STORAGE_ACCOUNT_KEY"]:
        if varname in os.environ:
            account = {}
            if "AZURE_STORAGE_ACCOUNT" in os.environ:
                account["account"] = os.environ["AZURE_STORAGE_ACCOUNT"]
            return {"_azure_auth": "sakey", "storage_account_key": os.environ[varname], **account}

    if "AZURE_STORAGE_CONNECTION_STRING" in os.environ:
        connection_data = {}
        # technically this should be parsed according to the rules in
        # https://www.connectionstrings.com/formating-rules-for-connection-strings/
        for part in os.environ["AZURE_STORAGE_CONNECTION_STRING"].split(";"):
            key, _, val = part.partition("=")
            connection_data[key.lower()] = val
        return {
            "_azure_auth": "sakey",
            "account": connection_data["accountname"],
            "storage_account_key": connection_data["accountkey"],
        }

    if "AZURE_CLIENT_ID" in os.environ:
        return {
            "_azure_auth": "svcact",
            "client_id": os.environ["AZURE_CLIENT_ID"],
            "client_secret": os.environ["AZURE_CLIENT_SECRET"],
            "tenant_id": os.environ["AZURE_TENANT_ID"],
        }

    # look for a refresh token in the az command line credentials
    # TODO: we could also try to use any found access tokens
    msal_tokens_path = os.path.expanduser("~/.azure/msal_token_cache.json")
    if os.path.exists(msal_tokens_path):
        with open(msal_tokens_path) as f:
            tokens = json.load(f)
            for token in tokens.get("RefreshToken", {}).values():
                if token["credential_type"] != "RefreshToken":
                    continue
                return {"_azure_auth": "refresh", "refresh_token": token["secret"]}

    access_tokens_path = os.path.expanduser("~/.azure/accessTokens.json")
    if os.path.exists(access_tokens_path):
        with open(access_tokens_path) as f:
            tokens = json.load(f)
            best_token = None
            for token in tokens:
                if "refreshToken" not in token:
                    continue
                creds = {"_azure_auth": "refresh", "refresh_token": token["refreshToken"]}
                if best_token is None:
                    best_token = creds
                else:
                    # expiresOn may be missing for tokens from service principals
                    if token.get("expiresOn", "") > best_token.get("expiresOn", ""):
                        best_token = creds
            if best_token is not None:
                return best_token

    raise RuntimeError(
        """Azure credentials not found, please do one of the following:

1) Log in with 'az login', boostedblob will use your default credentials to lookup your storage
   account key
2) Create an account with 'az ad sp create-for-rbac --name <name>' and
   individually set the 'AZURE_CLIENT_ID', 'AZURE_CLIENT_SECRET', and
   'AZURE_TENANT_ID' environment variables
3) Set the environment variable 'AZURE_STORAGE_ACCOUNT_KEY' to your storage account key which you
   can find by following this guide:
   https://docs.microsoft.com/en-us/azure/storage/common/storage-account-keys-manage
"""
    )


def load_stored_subscription_ids() -> List[str]:
    """Return a list of subscription ids from the local azure profile.

    The default subscription will appear first in the list.

    """
    default_profile_path = os.path.expanduser("~/.azure/azureProfile.json")
    if not os.path.exists(default_profile_path):
        return []

    with open(default_profile_path, "rb") as f:
        # this file has a UTF-8 BOM
        profile = json.loads(f.read().decode("utf-8-sig"))

    subscriptions = profile.get("subscriptions", [])
    subscriptions.sort(key=lambda x: x["isDefault"], reverse=True)
    return [sub["id"] for sub in subscriptions]


async def get_access_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, float]:
    account, container = cache_key

    now = time.time()
    creds = load_credentials()

    if creds["_azure_auth"] == "sakey":
        if "account" in creds:
            if creds["account"] != account:
                raise RuntimeError(
                    f"Provided storage account key for account '{creds['account']}' via "
                    f"environment variables, but needed credentials for account '{account}'"
                )
        auth = (SHARED_KEY, creds["storage_account_key"])
        if await can_access_account(account, container, auth):
            return (auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
        raise RuntimeError(
            f"Found storage account key, but it was unable to access storage account: '{account}'"
        )

    from .globals import config

    if creds["_azure_auth"] == "refresh":
        # we have a refresh token, convert it into an access token for this account
        req = create_access_token_request(
            creds=creds,
            scope=f"https://{account}.blob.core.windows.net/.default",
            success_codes=(200, 400),
        )

        async with req.execute() as resp:
            result = await resp.json()
            if resp.status == 400:
                if (
                    result["error"] == "invalid_grant"
                    and "AADSTS700082" in result["error_description"]
                ) or (
                    result["error"] == "interaction_required"
                    and "AADSTS50078" in result["error_description"]
                ):
                    raise RuntimeError(
                        "Your refresh token has expired, please run `az login` to refresh it"
                    )
                raise RuntimeError(
                    "Encountered an error when requesting an access token:\n"
                    f"{result['error']}: {result['error_description']}\n\n"
                    "Try running `az login`?"
                )

        auth = (OAUTH_TOKEN, result["access_token"])

        # for some azure accounts this access token does not work, check if it works
        if await can_access_account(account, container, auth):
            return (auth, now + float(result["expires_in"]))

        # it didn't work, fall back to getting the storage keys
        if config.storage_account_key_fallback:
            storage_account_key_auth = await get_storage_account_key(
                account=account, creds=creds, container_hint=container
            )
            if storage_account_key_auth is not None:
                return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)

    if creds["_azure_auth"] == "svcact":
        # we have a service principal, get an oauth token
        req = create_access_token_request(creds=creds, scope="https://storage.azure.com/.default")

        async with req.execute() as resp:
            result = await resp.json()
        auth = (OAUTH_TOKEN, result["access_token"])
        if await can_access_account(account, container, auth):
            return (auth, now + float(result["expires_in"]))

        # it didn't work, fall back to getting the storage keys
        if config.storage_account_key_fallback:
            storage_account_key_auth = await get_storage_account_key(
                account=account, creds=creds, container_hint=container
            )
            if storage_account_key_auth is not None:
                return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)

    raise RuntimeError(
        f"Could not find any credentials that grant access to storage account: '{account}'"
    )


async def can_access_account(account: str, container: Optional[str], auth: Tuple[str, str]) -> bool:
    from .request import Request, azurify_request

    if not container:
        # if a container isn't specified, check that we can list the storage account
        req = Request(
            method="GET",
            url=f"https://{account}.blob.core.windows.net",
            params={"comp": "list", "maxresults": "1"},
            success_codes=(200, 403),
        )
        req = await azurify_request(req, auth=auth)

        async with req.execute() as resp:
            if resp.status == 403:
                return False
            data = await resp.read()

        result = etree.fromstring(data)
        container = result.findtext("Containers/Container/Name")

        if container is None:
            # there are no containers in this storage account
            # we can't test if we can access this storage account or not, so presume we can
            return True
        # then also test that we can list that container. this is perhaps unnecessary...

    # https://myaccount.blob.core.windows.net/mycontainer?restype=container&comp=list
    req = Request(
        method="GET",
        url=f"https://{account}.blob.core.windows.net/{container}",
        params={"restype": "container", "comp": "list", "maxresults": "1"},
        success_codes=(200, 403),
    )
    req = await azurify_request(req, auth=auth)
    async with req.execute() as resp:
        return resp.status == 200


def create_access_token_request(
    creds: Mapping[str, str], scope: str, success_codes: Sequence[int] = (200,)
) -> Request:
    from .request import Request

    if creds["_azure_auth"] == "refresh":
        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-auth-code-flow#refresh-the-access-token
        data = {
            "grant_type": "refresh_token",
            "refresh_token": creds["refresh_token"],
            "scope": scope,
        }
        tenant_id = "common"
    elif creds["_azure_auth"] == "svcact":
        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-client-creds-grant-flow#first-case-access-token-request-with-a-shared-secret
        data = {
            "grant_type": "client_credentials",
            "client_id": creds["client_id"],
            "client_secret": creds["client_secret"],
            "scope": scope,
        }
        tenant_id = creds["tenant_id"]
    else:
        raise AssertionError
    return Request(
        url=f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
        method="POST",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data=urllib.parse.urlencode(data).encode("utf8"),
        success_codes=success_codes,
    )


async def get_storage_account_id_with_subscription(
    subscription_id: str, account: str, auth: Tuple[str, str]
) -> Optional[str]:
    from .request import Request, azurify_request

    req = Request(
        method="GET",
        url=f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Storage/storageAccounts",
        params={"api-version": "2019-04-01"},
        success_codes=(200, 401, 403, 429),
    )

    while True:
        req = await azurify_request(req, auth=auth)
        async with req.execute() as resp:
            if resp.status in (401, 403):
                # we aren't allowed to query this for this subscription, skip it
                return None
            if resp.status == 429:
                # these API endpoints have really low rate limits
                # note this error message only makes sense because the only code path that
                # currently calls get_storage_account_id_with_subscription is the storage
                # account key fallback
                raise RuntimeError(
                    "You do not have the correct IAM roles to access this location. "
                    "Check that you have Storage Blob Data Reader or Storage Blob Data Contributor "
                    "roles. You can run "
                    f"`az storage container list --auth-mode login --account-name {account}` "
                    "to confirm the missing role.\n"
                    "In this situation, boostedblob is capable of falling back to accessing the "
                    "storage account keys, but Azure has really low rate limits on these queries, "
                    "which are currently preventing us from doing so."
                )
            result = await resp.json()

        # search for the storage account
        account_id = next((obj["id"] for obj in result["value"] if obj["name"] == account), None)
        if account_id is not None:
            return account_id

        if "nextLink" not in result:
            break
        req = Request(method="GET", url=result["nextLink"], success_codes=(200, 401, 403))
    return None


async def get_storage_account_id(account: str, auth: Tuple[str, str]) -> Optional[str]:
    from .request import Request, azurify_request

    stored_subscription_ids = load_stored_subscription_ids()
    for subscription_id in stored_subscription_ids:
        storage_account_id = await get_storage_account_id_with_subscription(
            subscription_id, account, auth
        )
        if storage_account_id:
            return storage_account_id

    req = Request(
        method="GET",
        url="https://management.azure.com/subscriptions",
        params={"api-version": "2020-01-01"},
    )

    while True:
        req = await azurify_request(req, auth=auth)
        async with req.execute() as resp:
            result = await resp.json()
        subscription_ids = {item["subscriptionId"] for item in result["value"]}

        for subscription_id in subscription_ids - set(stored_subscription_ids):
            storage_account_id = await get_storage_account_id_with_subscription(
                subscription_id, account, auth
            )
            if storage_account_id:
                return storage_account_id

        if "nextLink" not in result:
            break
        req = Request(method="GET", url=result["nextLink"])
    return None


async def get_storage_account_key(
    account: str, creds: Mapping[str, Any], container_hint: Optional[str] = None
) -> Optional[Tuple[Any, float]]:
    from .request import Request, azurify_request

    # get an access token for the management service
    req = create_access_token_request(creds=creds, scope="https://management.azure.com/.default")

    async with req.execute() as resp:
        result = await resp.json()
    auth = (OAUTH_TOKEN, result["access_token"])

    storage_account_id = await get_storage_account_id(account, auth)
    if not storage_account_id:
        return None

    req = Request(
        method="POST",
        url=f"https://management.azure.com{storage_account_id}/listKeys",
        params={"api-version": "2019-04-01"},
    )
    req = await azurify_request(req, auth=auth)

    async with req.execute() as resp:
        result = await resp.json()
    for key in result["keys"]:
        if key["permissions"] == "FULL":
            storage_key_auth = (SHARED_KEY, key["value"])
            if await can_access_account(account, container_hint, storage_key_auth):
                return storage_key_auth
            raise RuntimeError(
                "Found storage account key, but it was unable to access "
                f"storage account: '{account}'"
            )
    raise RuntimeError(
        f"Storage account was found, but storage account keys were missing: '{account}'"
    )


def sign_request_with_shared_key(request: Request, key: str) -> str:
    # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-shared-key
    from yarl import URL

    # Figuring out that shared key authorisation sometimes breaks because aiohttp is a little over
    # clever about canonicalising URLs was not fun; preemptively canonicalise
    parsed_url = urllib.parse.urlparse(str(URL(request.url)))
    storage_account = parsed_url.netloc.split(".")[0]
    blob = parsed_url.path[1:]

    def canonicalized_resource() -> str:
        params_to_sign = sorted(f"{name.lower()}:{value}" for name, value in request.params.items())
        canonical_url = f"/{storage_account}/{blob}"
        return "\n".join([canonical_url] + params_to_sign)

    def canonicalized_headers() -> str:
        headers_to_sign = []
        for name, value in request.headers.items():
            canonical_name = name.lower()
            canonical_value = re.sub(r"\s+", " ", value).strip()
            if canonical_name.startswith("x-ms-"):
                headers_to_sign.append(f"{canonical_name}:{canonical_value}")
        return "\n".join(sorted(headers_to_sign))

    headers = request.headers
    content_length = headers.get("Content-Length", "")
    if request.data is not None:
        content_length = str(len(request.data))

    parts_to_sign = [
        request.method,
        headers.get("Content-Encoding", ""),
        headers.get("Content-Language", ""),
        content_length,
        headers.get("Content-MD5", ""),
        headers.get("Content-Type", ""),
        headers.get("Date", ""),
        headers.get("If-Modified-Since", ""),
        headers.get("If-Match", ""),
        headers.get("If-None-Match", ""),
        headers.get("If-Unmodified-Since", ""),
        headers.get("Range", ""),
        canonicalized_headers(),
        canonicalized_resource(),
    ]
    string_to_sign = "\n".join(parts_to_sign)

    signature = base64.b64encode(
        hmac.digest(base64.b64decode(key), string_to_sign.encode("utf8"), "sha256")
    ).decode("utf8")

    return f"SharedKey {storage_account}:{signature}"


async def get_sas_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, float]:
    from .globals import config
    from .request import Request, azurify_request

    auth = await config.azure_access_token_manager.get_token(key=cache_key)
    account, container = cache_key

    if auth[0] != OAUTH_TOKEN:
        cmd = (
            f"az storage container list --auth-mode login --account-name {account}"
            if container is None
            else (
                f"az storage blob list --auth-mode login --account-name {account} "
                f"--container {container}"
            )
        )
        raise RuntimeError(
            "Only OAuth tokens can be used to get SAS tokens. You should set the "
            "Blob Data Reader or Storage Blob Data Contributor IAM role. You can run "
            f"`{cmd}` to confirm that the missing role is the issue."
        )

    # https://docs.microsoft.com/en-us/rest/api/storageservices/create-user-delegation-sas
    now = datetime.datetime.utcnow()
    start = (now + datetime.timedelta(hours=-1)).strftime("%Y-%m-%dT%H:%M:%SZ")
    expiration = now + datetime.timedelta(days=6)
    expiry = expiration.strftime("%Y-%m-%dT%H:%M:%SZ")
    req = await azurify_request(
        Request(
            url=f"https://{account}.blob.core.windows.net/",
            method="POST",
            params=dict(restype="service", comp="userdelegationkey"),
            data={"KeyInfo": {"Start": start, "Expiry": expiry}},
            success_codes=(200, 403),
        ),
        auth=auth,
    )
    async with req.execute() as resp:
        if resp.status == 403:
            raise RuntimeError(
                f"You do not have permission to generate an SAS token for account {account}. "
                "Try setting the Storage Blob Delegator or Storage Blob Data Contributor IAM role "
                "at the account level."
            )
        data = await resp.read()

    result = etree.fromstring(data)
    user_delegation_key = {el.tag: el.text for el in result}
    return user_delegation_key, time.time() + AZURE_SAS_TOKEN_EXPIRATION_SECONDS


async def generate_signed_url(path: AzurePath) -> Tuple[str, datetime.datetime]:
    # https://docs.microsoft.com/en-us/rest/api/storageservices/delegate-access-with-shared-access-signature
    # https://docs.microsoft.com/en-us/rest/api/storageservices/create-user-delegation-sas
    # https://docs.microsoft.com/en-us/rest/api/storageservices/service-sas-examples
    from .globals import config

    key = await config.azure_sas_token_manager.get_token(key=(path.account, path.container or None))
    params = {
        "st": key["SignedStart"],
        "se": key["SignedExpiry"],
        "sks": key["SignedService"],
        "skt": key["SignedStart"],
        "ske": key["SignedExpiry"],
        "sktid": key["SignedTid"],
        "skoid": key["SignedOid"],
        # signed key version (param name not mentioned in docs)
        "skv": key["SignedVersion"],
        "sv": "2018-11-09",  # signed version
        "sr": "b",  # signed resource
        "sp": "r",  # signed permissions
        "sip": "",  # signed ip
        "si": "",  # signed identifier
        "spr": "https,http",  # signed http protocol
        "rscc": "",  # Cache-Control header
        "rscd": "",  # Content-Disposition header
        "rsce": "",  # Content-Encoding header
        "rscl": "",  # Content-Language header
        "rsct": "",  # Content-Type header
    }

    canonicalized_resource = f"/blob/{path.account}/{path.container}/{path.blob}"
    parts_to_sign = (
        params["sp"],
        params["st"],
        params["se"],
        canonicalized_resource,
        params["skoid"],
        params["sktid"],
        params["skt"],
        params["ske"],
        params["sks"],
        params["skv"],
        params["sip"],
        params["spr"],
        params["sv"],
        params["sr"],
        params["rscc"],
        params["rscd"],
        params["rsce"],
        params["rscl"],
        params["rsct"],
        # this is documented on a different page
        # https://docs.microsoft.com/en-us/rest/api/storageservices/create-service-sas#specifying-the-signed-identifier
        params["si"],
    )
    string_to_sign = "\n".join(parts_to_sign)
    params["sig"] = base64.b64encode(
        hmac.digest(base64.b64decode(key["Value"]), string_to_sign.encode("utf8"), "sha256")
    ).decode("utf8")
    query = urllib.parse.urlencode({k: v for k, v in params.items() if v != ""})

    expiry = datetime.datetime.strptime(key["SignedExpiry"], "%Y-%m-%dT%H:%M:%SZ")
    url = path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}")
    return url + "?" + query, expiry



from __future__ import annotations

import itertools
from typing import Any, Iterator, Optional, Tuple, Union

from .boost import (
    BoostExecutor,
    BoostUnderlying,
    OrderedMappingBoostable,
    UnorderedMappingBoostable,
)
from .globals import config
from .path import AzurePath, BasePath, CloudPath, GooglePath, LocalPath, getsize, pathdispatch
from .request import Request, azurify_request, execute_retrying_read, googlify_request

ByteRange = Tuple[int, int]
OptByteRange = Tuple[Optional[int], Optional[int]]

# ==============================
# read_byte_range
# ==============================


@pathdispatch
async def read_byte_range(path: Union[BasePath, str], byte_range: OptByteRange) -> bytes:
    """Read the content of ``path`` in the given byte range.

    :param path: The path to read from.
    :param byte_range: The byte range to read.
    :return: The relevant bytes.

    """
    raise ValueError(f"Unsupported path: {path}")


@read_byte_range.register  # type: ignore
async def _azure_read_byte_range(path: AzurePath, byte_range: OptByteRange) -> bytes:
    range_str = byte_range_to_str(byte_range)
    range_header = {"Range": range_str} if range_str is not None else {}
    success_codes = (206,) if range_header else (200,)
    request = await azurify_request(
        Request(
            method="GET",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            headers=range_header,
            success_codes=success_codes,
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    return await execute_retrying_read(request)


@read_byte_range.register  # type: ignore
async def _google_read_byte_range(path: GooglePath, byte_range: OptByteRange) -> bytes:
    range_str = byte_range_to_str(byte_range)
    range_header = {"Range": range_str} if range_str is not None else {}
    success_codes = (206,) if range_header else (200,)
    request = await googlify_request(
        Request(
            method="GET",
            url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o/{blob}"),
            params=dict(alt="media"),
            headers=range_header,
            success_codes=success_codes,
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    return await execute_retrying_read(request)


@read_byte_range.register  # type: ignore
async def _local_read_byte_range(path: LocalPath, byte_range: OptByteRange) -> bytes:
    with open(path, "rb") as f:
        start, end = byte_range
        if start is None:
            return f.read() if end is None else f.read(end)
        f.seek(start)
        return f.read() if end is None else f.read(end - start)


# ==============================
# read_single
# ==============================


@pathdispatch
async def read_single(path: Union[BasePath, str]) -> bytes:
    """Read the content of ``path``.

    :param path: The path to read from.

    """
    raise ValueError(f"Unsupported path: {path}")


@read_single.register  # type: ignore
async def _cloud_read_single(path: CloudPath) -> bytes:
    return await read_byte_range(path, (None, None))


@read_single.register  # type: ignore
async def _local_read_single(path: LocalPath) -> bytes:
    with open(path, "rb") as f:
        return f.read()


# ==============================
# read_stream
# ==============================


@pathdispatch
async def read_stream(
    path: Union[BasePath, str], executor: BoostExecutor, size: Optional[int] = None
) -> BoostUnderlying[bytes]:
    """Read the content of ``path``.

    :param path: The path to read from.
    :param executor: An executor.
    :param size: If specified, will save a network call.
    :return: The stream of bytes, chunking determined by ``config.chunk_size``.

    """
    raise ValueError(f"Unsupported path: {path}")


@read_stream.register  # type: ignore
async def _cloud_read_stream(
    path: CloudPath, executor: BoostExecutor, size: Optional[int] = None
) -> OrderedMappingBoostable[Any, bytes]:
    if size is None:
        size = await getsize(path)

    byte_ranges = itertools.zip_longest(
        range(0, size, config.chunk_size),
        range(config.chunk_size, size, config.chunk_size),
        fillvalue=size,
    )

    # Note that we purposefully don't do
    # https://docs.aiohttp.org/en/stable/client_quickstart.html#streaming-response-content
    # Doing that would stream data as we needed it, which is a little too lazy for our purposes
    chunks = executor.map_ordered(lambda byte_range: read_byte_range(path, byte_range), byte_ranges)
    return chunks


@read_stream.register  # type: ignore
async def _local_read_stream(
    path: LocalPath, executor: BoostExecutor, size: Optional[int] = None
) -> Iterator[bytes]:
    def iterator() -> Iterator[bytes]:
        with open(path, "rb") as f:
            while True:
                data = f.read(config.chunk_size)
                if not data:
                    return
                yield data

    return iterator()


# ==============================
# read_stream_unordered
# ==============================


@pathdispatch
async def read_stream_unordered(
    path: Union[CloudPath, str], executor: BoostExecutor, size: Optional[int] = None
) -> UnorderedMappingBoostable[Any, Tuple[bytes, ByteRange]]:
    assert isinstance(path, CloudPath)

    if size is None:
        size = await getsize(path)

    byte_ranges = itertools.zip_longest(
        range(0, size, config.chunk_size),
        range(config.chunk_size, size, config.chunk_size),
        fillvalue=size,
    )

    async def read_byte_range_wrapper(byte_range: ByteRange) -> Tuple[bytes, ByteRange]:
        chunk = await read_byte_range(path, byte_range)
        return (chunk, byte_range)

    chunks = executor.map_unordered(read_byte_range_wrapper, byte_ranges)
    return chunks


# ==============================
# helpers
# ==============================


def byte_range_to_str(byte_range: OptByteRange) -> Optional[str]:
    # https://docs.microsoft.com/en-us/rest/api/storageservices/specifying-the-range-header-for-blob-service-operations
    # https://cloud.google.com/storage/docs/xml-api/get-object-download
    # oddly range requests are not mentioned in JSON API, only in the XML API
    start, end = byte_range
    if start is not None and end is not None:
        return f"bytes={start}-{end-1}"
    if start is not None:
        return f"bytes={start}-"
    if end is not None:
        if end > 0:
            return f"bytes=0-{end-1}"
        # This form is not supported by Azure
        return f"bytes=-{-int(end)}"
    return None



import asyncio
import os
import shutil
from typing import AsyncIterator, Optional, Union

from .boost import BoostExecutor, consume
from .listing import glob_scandir, listtree
from .path import AzurePath, BasePath, CloudPath, GooglePath, LocalPath, isdir, isfile, pathdispatch
from .request import Request, azurify_request, googlify_request

# ==============================
# remove
# ==============================


@pathdispatch
async def remove(path: Union[BasePath, str]) -> BasePath:
    """Delete the file ``path``.

    :param path: The path to delete.

    """
    raise ValueError(f"Unsupported path: {path}")


@remove.register  # type: ignore
async def _azure_remove(path: AzurePath) -> AzurePath:
    request = await azurify_request(
        Request(
            method="DELETE",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            success_codes=(202,),
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    try:
        await request.execute_reponseless()
    except FileNotFoundError:
        # Note that this continues to allow the deletion of directory marker files
        if await isdir(path):
            raise IsADirectoryError(path) from None
        raise
    return path


@remove.register  # type: ignore
async def _google_remove(path: GooglePath) -> GooglePath:
    request = await googlify_request(
        Request(
            method="DELETE",
            url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o/{blob}"),
            success_codes=(204,),
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    try:
        await request.execute_reponseless()
    except FileNotFoundError:
        # Note that this continues to allow the deletion of directory marker files
        if await isdir(path):
            raise IsADirectoryError(path) from None
        raise
    return path


@remove.register  # type: ignore
async def _local_remove(path: LocalPath) -> LocalPath:
    os.remove(path)
    return path


# ==============================
# glob_remove
# ==============================


async def glob_remove(
    path: Union[BasePath, str], executor: BoostExecutor
) -> AsyncIterator[BasePath]:
    """Delete all files that match the glob ``path``.

    :param path: The glob to match against.
    :param executor: An executor.

    """
    async for subpath in executor.map_unordered(
        lambda x: remove(x.path), executor.eagerise(glob_scandir(path))
    ):
        yield subpath


# ==============================
# rmtree
# ==============================


async def rmtree_iterator(path: CloudPath, executor: BoostExecutor) -> AsyncIterator[BasePath]:
    """Delete the directory ``path``.

    Yields the deleted paths as they are deleted.

    :param path: The path to delete.
    :param executor: An executor.

    """
    # Note that this function almost works for LocalPath, except that we wouldn't remove empty
    # directories. To get this to work, we'd need a variant of scantree that also yields
    # directories, ensuring they're in the right order. This would also allow us to get rid of the
    # current directory file marker special casing.
    dirpath = path.ensure_directory_like()

    # listtree will not list the directory itself, so we need to remove the marker file if present
    async def remove_directory_marker() -> Optional[BasePath]:
        try:
            return await remove(dirpath)
        except (FileNotFoundError, IsADirectoryError, PermissionError):
            return None

    marker_task = asyncio.create_task(remove_directory_marker())

    try:
        async for subpath in executor.map_unordered(remove, executor.eagerise(listtree(dirpath))):
            yield subpath
    except FileNotFoundError:
        if await isfile(path):
            raise NotADirectoryError(path) from None
        raise

    marker = await marker_task
    if marker:
        yield marker


@pathdispatch
async def rmtree(path: Union[BasePath, str], executor: BoostExecutor) -> None:
    """Delete the directory ``path``.

    :param path: The path to delete.
    :param executor: An executor.

    """
    raise ValueError(f"Unsupported path: {path}")


@rmtree.register  # type: ignore
async def _cloud_rmtree(path: CloudPath, executor: BoostExecutor) -> None:
    await consume(rmtree_iterator(path, executor))


@rmtree.register  # type: ignore
async def _local_rmtree(path: LocalPath, executor: BoostExecutor) -> None:
    shutil.rmtree(path)



from __future__ import annotations

import asyncio
import base64
import datetime
import email.utils
import functools
import os
import urllib.parse
from dataclasses import dataclass
from typing import Any, Callable, Mapping, Optional, TypeVar, Union

from .request import Request, azure_page_iterator, azurify_request, googlify_request

T = TypeVar("T")

# ==============================
# Path classes
# ==============================


class BasePath:
    @staticmethod
    def from_str(path: str) -> BasePath:
        url = urllib.parse.urlparse(path)
        if url.scheme == "gs":
            return GooglePath.from_str(path)
        if url.scheme == "az" or (
            url.scheme == "https" and url.netloc.endswith(".blob.core.windows.net")
        ):
            return AzurePath.from_str(path)
        if url.scheme:
            raise ValueError(f"Invalid path '{path}'")
        return LocalPath(path)

    @property
    def name(self) -> str:
        """Returns the name of path, normalised to exclude any trailing slash."""
        raise NotImplementedError

    @property
    def parent(self: T) -> T:
        raise NotImplementedError

    def relative_to(self: T, other: T) -> str:
        raise NotImplementedError

    def is_relative_to(self: T, other: T) -> bool:
        try:
            self.relative_to(other)  # type: ignore[attr-defined]
            return True
        except ValueError:
            return False

    def is_directory_like(self) -> bool:
        raise NotImplementedError

    def ensure_directory_like(self: T) -> T:
        raise NotImplementedError

    def __truediv__(self: T, relative_path: str) -> T:
        raise NotImplementedError


@dataclass(frozen=True)
class LocalPath(BasePath):
    path: str

    @staticmethod
    def from_str(path: str) -> BasePath:
        url = urllib.parse.urlparse(path)
        if url.scheme:
            raise ValueError(f"Invalid path '{path}'")
        return LocalPath(path)

    @property
    def name(self) -> str:
        return os.path.basename(_strip_slash(self.path))

    @property
    def parent(self) -> LocalPath:
        parent = os.path.dirname(_strip_slash(self.path))
        if not parent:
            parent = "/" if os.path.isabs(self.path) else "."
        return LocalPath(parent)

    def relative_to(self, other: LocalPath) -> str:
        if not isinstance(other, LocalPath):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        other = other.ensure_directory_like()
        other_path = other.path
        if other_path.startswith("./") and not self.path.startswith("./"):
            other_path = other_path[len("./") :]
        if not self.path.startswith(other_path):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        return self.path[len(other_path) :]

    def is_directory_like(self) -> bool:
        return not self.path or self.path.endswith("/")

    def ensure_directory_like(self) -> LocalPath:
        return self if self.is_directory_like() else LocalPath(self.path + "/")

    def abspath(self) -> LocalPath:
        return LocalPath(os.path.abspath(self.path))

    def __truediv__(self, relative_path: str) -> LocalPath:
        return LocalPath(os.path.join(self.path, relative_path))

    def __str__(self) -> str:
        return self.path

    def __fspath__(self) -> str:
        return self.path


class CloudPath(BasePath):
    pass


@dataclass(frozen=True)
class AzurePath(CloudPath):
    account: str
    container: str
    blob: str

    @staticmethod
    def from_str(url: str) -> AzurePath:
        parsed_url = urllib.parse.urlparse(url)
        if parsed_url.scheme == "https":
            account, host = parsed_url.netloc.split(".", maxsplit=1)
            if host != "blob.core.windows.net":
                raise ValueError(f"Invalid URL '{url}'; unexpected host '{host}'")
        elif parsed_url.scheme == "az":
            account = parsed_url.netloc
        else:
            raise ValueError(f"Invalid URL '{url}'; expected 'https' or 'az' scheme")

        # split the unparsed URL
        parts = url.split("/", maxsplit=4)
        container = parts[3] if len(parts) >= 4 else ""
        blob = parts[4] if len(parts) >= 5 else ""
        return AzurePath(account=account, container=container, blob=blob)

    @property
    def name(self) -> str:
        return os.path.basename(_strip_slash(self.blob)) if self.blob else self.container

    @property
    def parent(self) -> AzurePath:
        return AzurePath(self.account, self.container, os.path.dirname(_strip_slash(self.blob)))

    def relative_to(self, other: AzurePath) -> str:
        if not isinstance(other, AzurePath):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        other = other.ensure_directory_like()
        if (
            self.account != other.account
            or self.container != other.container
            or not self.blob.startswith(other.blob)
        ):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        return self.blob[len(other.blob) :]

    def is_directory_like(self) -> bool:
        return not self.blob or self.blob.endswith("/")

    def ensure_directory_like(self) -> AzurePath:
        return (
            self
            if self.is_directory_like()
            else AzurePath(self.account, self.container, self.blob + "/")
        )

    def __truediv__(self, relative_path: str) -> AzurePath:
        return AzurePath(self.account, self.container, os.path.join(self.blob, relative_path))

    def __str__(self) -> str:
        return self.to_az_str()

    def to_https_str(self) -> str:
        return f"https://{self.account}.blob.core.windows.net/{self.container}/{self.blob}"

    def to_az_str(self) -> str:
        return f"az://{self.account}/{self.container}/{self.blob}"

    def format_url(self, template: str) -> str:
        return url_format(template, account=self.account, container=self.container, blob=self.blob)


@dataclass(frozen=True)
class GooglePath(CloudPath):
    bucket: str
    blob: str

    @staticmethod
    def from_str(url: str) -> GooglePath:
        parsed_url = urllib.parse.urlparse(url)
        if parsed_url.scheme != "gs":
            raise ValueError(f"Invalid URL '{url}'; expected 'gs' scheme")

        # split the unparsed URL
        parts = url.split("/", maxsplit=3)
        blob = parts[3] if len(parts) >= 4 else ""
        return GooglePath(bucket=parsed_url.netloc, blob=blob)

    @property
    def name(self) -> str:
        return os.path.basename(_strip_slash(self.blob)) if self.blob else self.bucket

    @property
    def parent(self) -> GooglePath:
        return GooglePath(self.bucket, os.path.dirname(_strip_slash(self.blob)))

    def relative_to(self, other: GooglePath) -> str:
        if not isinstance(other, GooglePath):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        other = other.ensure_directory_like()
        if self.bucket != other.bucket or not self.blob.startswith(other.blob):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        return self.blob[len(other.blob) :]

    def is_directory_like(self) -> bool:
        return not self.blob or self.blob.endswith("/")

    def ensure_directory_like(self) -> GooglePath:
        return self if self.is_directory_like() else GooglePath(self.bucket, self.blob + "/")

    def __truediv__(self, relative_path: str) -> GooglePath:
        return GooglePath(self.bucket, os.path.join(self.blob, relative_path))

    def __str__(self) -> str:
        return f"gs://{self.bucket}/{self.blob}"

    def format_url(self, template: str) -> str:
        return url_format(template, bucket=self.bucket, blob=self.blob)


# ==============================
# pathdispatch
# ==============================

F = TypeVar("F", bound=Callable[..., Any])


def pathdispatch(fn: F) -> F:
    """Implements the singledispatch pattern for paths.

    The extra work this does over singledispatch is that if the path argument is a str, it's
    automatically converted to the correct BasePath object and re-dispatched.

    Typing this function as we've done allows us to exactly type usage, but unfortunately it
    causes type checkers to complain about "no attribute register"

    """
    ret: F = functools.singledispatch(fn)  # type: ignore[assignment]

    @ret.register  # type: ignore
    def strdispatch(path: str, *args: Any, **kwargs: Any) -> Any:
        return ret(BasePath.from_str(path), *args, **kwargs)

    return ret


# ==============================
# stat
# ==============================


class Stat:
    size: int
    mtime: float
    ctime: float
    md5: Optional[str]
    version: Optional[str]


class AzureStat(Stat):
    __slots__ = (
        "size",
        "_mtime",
        "_raw_mtime",
        "_ctime",
        "_raw_ctime",
        "_md5",
        "_raw_md5",
        "version",
    )

    def __init__(self, item: Mapping[str, Any]) -> None:
        self.size = int(item["Content-Length"])
        self.version = item["Etag"]
        self._raw_ctime = (
            item["Creation-Time"] if "Creation-Time" in item else item["x-ms-creation-time"]
        )
        self._ctime: Optional[float] = None
        self._raw_mtime = item["Last-Modified"]
        self._mtime: Optional[float] = None
        self._raw_md5 = item.get("Content-MD5")
        self._md5: Optional[str] = ""

    @property
    def ctime(self) -> float:  # type: ignore[override]
        if self._ctime is None:
            self._ctime = _azure_parse_timestamp(self._raw_ctime)
        return self._ctime

    @property
    def mtime(self) -> float:  # type: ignore[override]
        if self._mtime is None:
            self._mtime = _azure_parse_timestamp(self._raw_mtime)
        return self._mtime

    @property
    def md5(self) -> Optional[str]:  # type: ignore[override]
        if self._md5 == "":
            self._md5 = _azure_get_md5(self._raw_md5)
        return self._md5

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stat):
            return False
        return (
            self.size == other.size
            and self.mtime == other.mtime
            and self.ctime == other.ctime
            and self.md5 == other.md5
            and self.version == other.version
        )


class GoogleStat(Stat):
    def __init__(self, item: Mapping[str, Any]) -> None:
        self.size = int(item["size"])
        self.mtime = _google_parse_timestamp(item["updated"])
        self.ctime = _google_parse_timestamp(item["timeCreated"])
        self.md5 = _google_get_md5(item)
        self.version = item["generation"]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stat):
            return False
        return (
            self.size == other.size
            and self.mtime == other.mtime
            and self.ctime == other.ctime
            and self.md5 == other.md5
            and self.version == other.version
        )


class LocalStat(Stat):
    def __init__(self, item: os.stat_result) -> None:
        self.size = item.st_size
        self.mtime = item.st_mtime
        self.ctime = item.st_ctime
        self.md5 = None
        self.version = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stat):
            return False
        return (
            self.size == other.size
            and self.mtime == other.mtime
            and self.ctime == other.ctime
            and self.md5 == other.md5
            and self.version == other.version
        )


@pathdispatch
async def stat(path: Union[BasePath, str]) -> Stat:
    raise ValueError(f"Unsupported path: {path}")


@stat.register  # type: ignore
async def _azure_stat(path: AzurePath) -> AzureStat:
    if not path.blob:
        raise FileNotFoundError(path)
    request = await azurify_request(
        Request(
            method="HEAD",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    async with request.execute() as resp:
        return AzureStat(resp.headers)


@stat.register  # type: ignore
async def _google_stat(path: GooglePath) -> GoogleStat:
    if not path.blob:
        raise FileNotFoundError(path)
    request = await googlify_request(
        Request(
            method="GET",
            url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o/{blob}"),
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    async with request.execute() as resp:
        result = await resp.json()
        return GoogleStat(result)


@stat.register  # type: ignore
async def _local_stat(path: LocalPath) -> LocalStat:
    return LocalStat(os.stat(path))


# ==============================
# getsize
# ==============================


@pathdispatch
async def getsize(path: Union[BasePath, str]) -> int:
    s = await stat(path)
    return s.size


# ==============================
# isdir
# ==============================


@pathdispatch
async def isdir(path: Union[BasePath, str], raise_on_missing: bool = False) -> bool:
    """Check whether ``path`` is a directory.

    :param path: The path that could be a directory.
    :param raise_on_missing: If True, raise FileNotFoundError if ``path`` does not exist. Otherwise,
        return False.

    """
    raise ValueError(f"Unsupported path: {path}")


@isdir.register  # type: ignore
async def _azure_isdir(path: AzurePath) -> bool:
    try:
        if path.blob:
            prefix: str = path.blob
            if not prefix.endswith("/"):
                prefix += "/"

            it = azure_page_iterator(
                Request(
                    method="GET",
                    url=path.format_url("https://{account}.blob.core.windows.net/{container}"),
                    params=dict(
                        comp="list",
                        restype="container",
                        prefix=prefix,
                        delimiter="/",
                        maxresults="1",
                    ),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            async for result in it:
                blobs = result.find("Blobs")
                assert blobs is not None
                if len(blobs):
                    return blobs.find("Blob") is not None or blobs.find("BlobPrefix") is not None
            return False
        else:
            request = await azurify_request(
                Request(
                    method="GET",
                    url=path.format_url("https://{account}.blob.core.windows.net/{container}"),
                    params=dict(restype="container"),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            await request.execute_reponseless()
            return True
    except FileNotFoundError:
        # execute might raise FileNotFoundError if storage account doesn't exist
        return False


@isdir.register  # type: ignore
async def _google_isdir(path: GooglePath) -> bool:
    try:
        if path.blob:
            prefix: str = path.blob
            if not prefix.endswith("/"):
                prefix += "/"
            request = await googlify_request(
                Request(
                    method="GET",
                    url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o"),
                    params=dict(prefix=prefix, delimiter="/", maxResults="1"),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            async with request.execute() as resp:
                result = await resp.json()
                return "items" in result or "prefixes" in result
        else:
            request = await googlify_request(
                Request(
                    method="GET",
                    url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}"),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            async with request.execute() as resp:
                return True
    except FileNotFoundError:
        return False


@isdir.register  # type: ignore
async def _local_isdir(path: LocalPath) -> bool:
    return os.path.isdir(path)


# ==============================
# isfile
# ==============================


@pathdispatch
async def isfile(path: Union[BasePath, str]) -> bool:
    raise ValueError(f"Unsupported path: {path}")


@isfile.register  # type: ignore
async def _cloud_isfile(path: CloudPath) -> bool:
    try:
        await stat(path)
        return True
    except FileNotFoundError:
        return False


@isfile.register  # type: ignore
async def _local_isfile(path: LocalPath) -> bool:
    return os.path.isfile(path)


# ==============================
# exists
# ==============================


@pathdispatch
async def exists(path: Union[BasePath, str]) -> int:
    raise ValueError(f"Unsupported path: {path}")


@exists.register  # type: ignore
async def _cloud_exists(path: CloudPath) -> int:
    # TODO: this is two network requests, make it one
    for fut in asyncio.as_completed([isfile(path), isdir(path)]):
        if await fut:
            return True
    return False


@exists.register  # type: ignore
async def _local_exists(path: LocalPath) -> int:
    return os.path.exists(path)


# ==============================
# helpers
# ==============================


def url_format(template: str, **data: Any) -> str:
    escaped_data = {k: urllib.parse.quote(v, safe="") for k, v in data.items()}
    return template.format(**escaped_data)


def _strip_slash(path: str) -> str:
    return path[:-1] if path.endswith("/") else path


def _azure_parse_timestamp(text: str) -> float:
    # A little more than 2x faster than:
    # strptime(text.replace("GMT", "Z"), "%a, %d %b %Y %H:%M:%S %z").timestamp()
    return email.utils.parsedate_to_datetime(text).timestamp()


def _azure_get_md5(content_md5: Optional[str]) -> Optional[str]:
    if content_md5 is None:
        return None
    return base64.b64decode(content_md5).hex()


def _google_parse_timestamp(text: str) -> float:
    return datetime.datetime.strptime(text, "%Y-%m-%dT%H:%M:%S.%f%z").timestamp()


def _google_get_md5(metadata: Mapping[str, Any]) -> Optional[str]:
    if "md5Hash" in metadata:
        return base64.b64decode(metadata["md5Hash"]).hex()

    if "metadata" in metadata and "md5" in metadata["metadata"]:
        # fallback to our custom hash if this is a composite object that is lacking the
        # md5Hash field
        return metadata["metadata"]["md5"]

    return None



"""

Experimental code to recover deleted or overwritten data in Azure.

This has been used and should work, but isn't tested to the degree something like this should be.

"""
import datetime
import urllib.parse
from typing import Any, DefaultDict, Dict, List, Union

from .boost import BoostExecutor
from .path import AzurePath
from .request import Request, azure_page_iterator, azurify_request


async def _listtree_versions_snapshots(
    prefix: Union[str, AzurePath]
) -> Dict[AzurePath, List[Dict[str, Any]]]:
    if isinstance(prefix, str):
        prefix = AzurePath.from_str(prefix)

    # Azure has both "snapshots" and "versions"
    # They are very similar, because Azure likes to keep things confusing

    # We assume one of the following two situations:
    # 1) Soft-delete is enabled, but versioning is not (preferred)
    # 2) Soft-delete and versioning is enabled
    # https://docs.microsoft.com/en-us/azure/storage/blobs/soft-delete-blob-overview
    # https://docs.microsoft.com/en-us/azure/storage/blobs/soft-delete-blob-overview#blob-soft-delete-and-versioning

    it = azure_page_iterator(
        Request(
            method="GET",
            url=prefix.format_url("https://{account}.blob.core.windows.net/{container}"),
            params=dict(
                comp="list",
                restype="container",
                prefix=prefix.blob,
                include="deleted,snapshots,versions",
            ),
        )
    )

    results = DefaultDict[AzurePath, List[Dict[str, Any]]](list)
    async for result in it:
        blobs = result.find("Blobs")
        assert blobs is not None
        for b in blobs.iterfind("Blob"):
            name: str = b.findtext("Name")  # type: ignore
            b_dict = {el.tag: el.text for el in b}
            results[AzurePath(prefix.account, prefix.container, name)].append(b_dict)
    return results


async def _undelete(path: AzurePath) -> AzurePath:
    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            params=dict(comp="undelete"),
            success_codes=(200,),
        )
    )
    await request.execute_reponseless()
    return path


async def _promote_candidate(path: AzurePath, candidate: Dict[str, Any]) -> None:
    if "Snapshot" in candidate:
        source = "snapshot=" + urllib.parse.quote(candidate["Snapshot"])
    elif "VersionId" in candidate:
        source = "versionid=" + urllib.parse.quote(candidate["VersionId"])
    else:
        raise AssertionError
    url = path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}")
    request = await azurify_request(
        Request(
            method="PUT",
            url=url,
            headers={"x-ms-copy-source": url + "?" + source},
            success_codes=(202,),
        )
    )
    async with request.execute() as resp:
        copy_status = resp.headers["x-ms-copy-status"]
        assert copy_status == "success"


async def _delete_snapshot(path: AzurePath, snapshot: str) -> None:
    request = await azurify_request(
        Request(
            method="DELETE",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            params={"snapshot": snapshot},
            success_codes=(202,),
        )
    )
    await request.execute_reponseless()


async def _recover_snapshot(
    path: AzurePath, candidate: Dict[str, Any], alternatives: List[Dict[str, Any]]
) -> None:
    # https://docs.microsoft.com/en-us/azure/storage/blobs/soft-delete-blob-overview#restoring-soft-deleted-objects
    await _undelete(path)
    if "Snapshot" in candidate:
        await _promote_candidate(path, candidate)
    # undelete will undelete all snapshots, so re-delete all previously deleted snapshots
    # (yes, yes, awaiting in a loop)
    if "Snapshot" in candidate and candidate.get("Deleted") == "true":
        await _delete_snapshot(path, candidate["Snapshot"])
    for b in alternatives:
        if "Snapshot" in b and b.get("Deleted") == "true":
            await _delete_snapshot(path, b["Snapshot"])


async def _determine_candidate_and_recover(
    path: AzurePath, restore_ts: str, versions_snapshots: List[Dict[str, Any]], dry_run: bool = True
) -> str:
    if any("VersionId" in b for b in versions_snapshots):
        # If using versioning, I think all blobs should have versions
        assert all("VersionId" in b for b in versions_snapshots)
        versions = versions_snapshots

        # VersionId marks the time that version of the blob was created
        versions.sort(key=lambda b: b["VersionId"], reverse=True)
        candidate_index = next(
            (i for i in range(len(versions)) if versions[i]["VersionId"] <= restore_ts), None
        )
        should_recover = (
            candidate_index is not None
            and versions[candidate_index].get("IsCurrentVersion") != "true"
        )

        if should_recover:
            assert candidate_index is not None
            candidate_desc = repr(versions[candidate_index]["VersionId"])
            desc = f"Recovering {path} to {candidate_desc}."
        else:
            desc = f"Not attempting to recover {path}."

        alternatives = [b for i, b in enumerate(versions) if i != candidate_index]
        alternatives_desc = ", ".join(repr(b["VersionId"]) for b in alternatives)
        if alternatives:
            desc += f" Alternatives include {alternatives_desc}."

        if dry_run or not should_recover:
            return desc

        assert candidate_index is not None
        await _promote_candidate(path, versions[candidate_index])
        return desc
    else:
        # If using snapshotting, all except one of them should be snapshots
        assert sum(1 for b in versions_snapshots if "Snapshot" not in b) == 1
        snapshots = versions_snapshots

        # Snapshot marks the time that version of the blob was obsoleted (I think)
        # Note the logic and sorting is different from the versioning case
        snapshots.sort(key=lambda b: b.get("Snapshot", "9999-99-99"))

        candidate_index = next(
            i
            for i in range(len(snapshots))
            if snapshots[i].get("Snapshot", "9999-99-99") >= restore_ts
        )
        should_recover = (
            "Snapshot" in snapshots[candidate_index]
            or snapshots[candidate_index].get("Deleted") == "true"
        )

        def _to_ts_desc(b: Dict[str, Any]) -> str:
            if "Snapshot" in b:
                return b["Snapshot"]
            assert candidate_index is not None
            deleted_time = snapshots[candidate_index]["Properties"]["DeletedTime"]
            return datetime.datetime.strptime(deleted_time, "%a, %d %b %Y %H:%M:%S GMT").strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )

        if should_recover:
            candidate_desc = repr(_to_ts_desc(snapshots[candidate_index]))
            desc = f"Recovering {path} to {candidate_desc}."
        else:
            desc = f"Not attempting to recover {path}."

        alternatives = [b for i, b in enumerate(snapshots) if i != candidate_index]
        alternatives_desc = ", ".join(repr(_to_ts_desc(b)) for b in alternatives)
        if alternatives:
            desc += f" Alternatives include {alternatives_desc}."

        if dry_run or not should_recover:
            return desc

        assert candidate_index is not None
        await _recover_snapshot(path, snapshots[candidate_index], alternatives)
        return desc


async def _recoverprefix(
    prefix: Union[str, AzurePath],
    restore_dt: Union[str, datetime.datetime],
    executor: BoostExecutor,
    dry_run: bool = True,
) -> None:
    """Attempt to recover blobs with a prefix to prevous snapshots or versions.

    For each blob, attempts to restore to the most recent version or snapshot that is older than or
    equal to restore_dt. Note that this isn't a point-in-time restore, e.g. it won't delete
    blobs that were added after restore_dt, or it may restore blobs that were actually only deleted
    after restore_dt.

    Please try to avoid blindly trusting this!
    """
    if restore_dt == "current":
        # mostly just useful for seeing the snapshots / versions that exist
        restore_dt = datetime.datetime.now(datetime.timezone.utc)
    if isinstance(restore_dt, str):
        restore_ts = restore_dt
    else:
        assert restore_dt.tzinfo is datetime.timezone.utc
        restore_ts = restore_dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    versions_snapshots = await _listtree_versions_snapshots(prefix)
    async for desc in executor.map_unordered(
        lambda p_vs: _determine_candidate_and_recover(
            path=p_vs[0], restore_ts=restore_ts, versions_snapshots=p_vs[1], dry_run=dry_run
        ),
        iter(versions_snapshots.items()),
    ):
        print(desc)

    print()
    print("This code is experimental, please verify that it actually does what you want.")
    if dry_run:
        print("Dry run, no actions actually taken.")



import asyncio
import os
import re
import sys
from dataclasses import dataclass
from typing import AsyncIterator, Iterator, List, Optional, Tuple, Union

from .boost import BoostExecutor
from .copying import copyfile
from .delete import remove
from .listing import DirEntry, scantree
from .path import BasePath, LocalPath, Stat


@dataclass(frozen=True)
class Action:
    relpath: str


@dataclass(frozen=True)
class CopyAction(Action):
    size: Optional[int]


@dataclass(frozen=True)
class DeleteAction(Action):
    pass


async def sync_action_iterator(
    src: BasePath, dst: BasePath, exclude: Optional[str] = None
) -> Iterator[Action]:
    """Yields the actions to take to sync the tree rooted at ``src`` to ``dst``.

    :param src: The root of the tree to copy from.
    :param dst: The root of the tree to copy to.

    """
    if isinstance(src, LocalPath):
        src = LocalPath(os.path.abspath(src.path))
    if isinstance(dst, LocalPath):
        dst = LocalPath(os.path.abspath(dst.path))

    try:
        exclude_pattern = re.compile(exclude) if exclude else None
    except re.error as e:
        raise ValueError(
            f"Failed to compile exclude pattern {repr(exclude)}: {e}\n"
            "Hint: exclude patterns should be Python regular expressions, not globs."
        ) from None

    async def collect_tree(tree: BasePath) -> List[Tuple[str, DirEntry]]:
        try:
            # Note that if you create marker files to mark directories, scantree will return
            # those. Filter by is_file to avoid syncing these files that represent directories.
            files = [(p.path.relative_to(tree), p) async for p in scantree(tree) if p.is_file]
            return [p for p in files if exclude_pattern is None or not exclude_pattern.search(p[0])]
        except FileNotFoundError:
            return []

    # We block on tree collection, rather than streaming actions, for fear that the list
    # operations might start to reflect our changes, causing weird things to happen.
    src_files, dst_files = await asyncio.gather(collect_tree(src), collect_tree(dst))
    return sync_files_action_iterator(src_files, dst_files)


def sync_files_action_iterator(
    src_files: List[Tuple[str, DirEntry]], dst_files: List[Tuple[str, DirEntry]]
) -> Iterator[Action]:
    src_files.sort(key=lambda p: p[0])
    dst_files.sort(key=lambda p: p[0])

    i = 0
    j = 0
    while i < len(src_files) or j < len(dst_files):
        if i < len(src_files) and (j == len(dst_files) or src_files[i][0] < dst_files[j][0]):
            src_stat = src_files[i][1].stat
            size = src_stat.size if src_stat else None
            yield CopyAction(src_files[i][0], size)
            i += 1
            continue
        if j < len(dst_files) and (i == len(src_files) or src_files[i][0] > dst_files[j][0]):
            yield DeleteAction(dst_files[j][0])
            j += 1
            continue
        if src_files[i][0] == dst_files[j][0]:
            if should_copy(src_files[i][1].stat, dst_files[j][1].stat):
                src_stat = src_files[i][1].stat
                size = src_stat.size if src_stat else None
                yield CopyAction(src_files[i][0], size)
            i += 1
            j += 1
            continue
        raise AssertionError


# ==============================
# sync
# ==============================


async def sync(
    src: Union[str, BasePath],
    dst: Union[str, BasePath],
    executor: BoostExecutor,
    delete: bool = False,
    exclude: Optional[str] = None,
) -> AsyncIterator[BasePath]:
    """Syncs the tree rooted at ``src`` to ``dst``.

    Yields the paths as they are copied or deleted.

    :param src: The root of the tree to sync from.
    :param dst: The root of the tree to sync to.
    :param executor: An executor.
    :param delete: Whether to delete files present in the destination but missing in the source.

    """
    src_obj = src if isinstance(src, BasePath) else BasePath.from_str(src)
    src_obj = src_obj.ensure_directory_like()
    dst_obj = dst if isinstance(dst, BasePath) else BasePath.from_str(dst)
    dst_obj = dst_obj.ensure_directory_like()

    if src_obj.is_relative_to(dst_obj) or dst_obj.is_relative_to(src_obj):
        raise ValueError("Cannot sync overlapping directories")

    async def copy_wrapper(relpath: str, size: Optional[int]) -> Optional[BasePath]:
        src_file = src_obj / relpath
        dst_file = dst_obj / relpath
        try:
            await copyfile(src_file, dst_file, executor, size=size, overwrite=True)
            return src_file
        except FileNotFoundError as e:
            print(
                "[boostedblob] File disappeared while syncing, ignoring. Likely due to "
                f"concurrent deletion: {e}",
                file=sys.stderr,
            )
        return None

    async def delete_wrapper(relpath: str) -> BasePath:
        dst_file = dst_obj / relpath
        await remove(dst_file)
        return dst_file

    async def action_wrapper(action: Action) -> Optional[BasePath]:
        if isinstance(action, CopyAction):
            return await copy_wrapper(action.relpath, action.size)
        if isinstance(action, DeleteAction):
            if delete:
                return await delete_wrapper(action.relpath)
        return None

    actions = executor.map_unordered(
        action_wrapper, await sync_action_iterator(src_obj, dst_obj, exclude=exclude)
    )
    async for path in actions:
        if path is not None:
            yield path


def should_copy(src_stat: Optional[Stat], dst_stat: Optional[Stat]) -> bool:
    if src_stat is None and dst_stat is None:
        return False
    if src_stat is None or dst_stat is None:
        # We'll only run into this if a cloud directory has a file marker. For most cases in which
        # we're asked to overwrite directories with files or vice versa, we'll either get an
        # IsADirectoryError / NotADirectoryError from local directories or no error because
        # directories are an illusion in cloud storage.
        raise IsADirectoryError(
            "Sync does not support overwriting directories with files or vice versa. "
            "Please rmtree your destination and try again"
        )
    if src_stat.size != dst_stat.size:
        return True
    if src_stat.md5 and dst_stat.md5:
        return src_stat.md5 != dst_stat.md5
    # Round mtime, since different stores have different precisions
    if int(src_stat.mtime) >= int(dst_stat.mtime):
        return True
    # If hashes are unavailable, sizes are the same and the dst file has a newer mtime
    # than the src file, we take our chances.
    return False



from typing import TYPE_CHECKING, Any, Mapping, Optional

if TYPE_CHECKING:
    import xml.etree.ElementTree as _etree

    etree = _etree
else:
    from lxml import etree


def dict_to_xml(dictionary: Mapping[str, Any]) -> bytes:
    if len(dictionary) != 1:
        raise ValueError("Document must have exactly one root")

    def inner(parent_key: str, value: Any, tree: Optional[etree.Element]) -> None:
        # Child is responsible for adding the parent tag to tree
        if isinstance(value, dict):
            current = root if tree is None else etree.SubElement(tree, parent_key)
            for k, v in value.items():
                inner(k, v, current)
        elif isinstance(value, list):
            for v in value:
                inner(parent_key, v, tree)
        else:
            assert tree is not None
            current = etree.SubElement(tree, parent_key)
            current.text = str(value)

    root_key, root_value = next(iter(dictionary.items()))
    root = etree.Element(root_key)
    inner(root_key, root_value, None)

    return etree.tostring(root, encoding="utf-8", xml_declaration=True)  # type: ignore[call-overload]
